// Implementation of elliptic curves method (ecm) V1.08
// written by Thorsten Reinecke, 23.06.99
// last change: 30.12.99

// Bemerkung:
// Die ursprngliche Fassung der Weierstrass-Form y^2=x^2+a*x+b
// wurde in dieser Version zugunsten der Montgomery-Form B*y^2=x^3+A*x^2+x
// modifiziert.

// soweit erste Tests ergeben, scheint das Modul zu laufen:

// Phase 1 wird nun in der Form (x::z) durchlaufen
// erst fr Phase 2 wird in die Form (x:y:1) konvertiert.

// Phase 2 (in improved continuation) ist nun deutlich schneller.
// Hier ist die Form (x:y:1) ideal, da die Punkte mit mit nur
// einer Subtraktion+Multiplikation aufgesammelt werden knnen und
// nur eine aufwendige Berechnung (gcd zum Faktortest + Berechnung
// der nchsten Intervallgrenze) pro Intervall anfllt.
// Die Form (x::z) wre hier aufwendiger.

// Phase 2 (fft continuation) ist asymptotisch schneller
// als die improved continuation. Das liegt daran, da die innerhalb
// der Suchintervalle liegenden Punkte als Nullstellen eines Polynoms
// interpretiert werden knnen. Wenn nun viele Suchintervalle berprft
// werden mssen, knnen asymptotisch schnelle Verfahren angewandt werden,
// die Polynome in mehreren Punkten zugleich auswerten.
// Die aktuelle Implementierung der fft-continuation ist noch nicht ausgereift,
// zeigt sich aber bereits jetzt fr grere Suchrume als berlegen.
// Nachteil: hoher Speicherbedarf


// #define DEBUGMODE /* funktioniert nur im LANGSAM-Modus ohne NRESIDUE! */
// #define LANGSAM /* in allen Phasen in der Form (x:y:1) rechnen */


// #define NRESIDUE /* use "modular_mult.cc" instead of mpz_mod (if possible) */
//#define ECM_FFT_CONTINUATION /* use fft-continuation instead of standard improved continuation */


#ifdef NRESIDUE
#include "modular_mult.cc"
#define modulo(a,b,n) NResidue.redc(a,b)
#else
#define modulo(a,b,n) mpz_mod(a,b,n)
#endif

#ifdef ECM_FFT_CONTINUATION
#include "polynomial.cc"
#endif


class elliptic_curves
{
 private:
  mpz_t a,b; // auer in go nur lesend verwendbar!
  mpz_t h,k,m; // allgemein verwendbar (zwischen Aufrufen vernderbar)!
  mpz_t x3,y3; // nur in add, mul2 verwendbar!
  mpz_t xh_mul, yh_mul; // nur in mul verwendbar!
  int sigma; // ber sigma wird die Kurve ausgewhlt/definiert
#ifdef NRESIDUE
  CN_Residue NResidue;
#endif
 public:
  inline elliptic_curves() /* Konstruktor */
#ifdef NRESIDUE
  : NResidue()
 /* Konstruktor aufrufen,
    NResidue wird spter mit n initialisiert.
    n ist DIE globale Variable (die faktorisiert werden soll)
    zugegeben: es wre besser, n trotzdem als Parameter allen Klassen
    zu bergeben, aber ich will's jetzt nicht mehr ndern, da
    dann auch die Rckgabe der Faktoren und die Division n/FAKTOR
    durchgefhrt werden mssen... 
 */
#endif   
    {
      mpz_init(a); mpz_init(b);
      mpz_init(h); mpz_init(k); mpz_init(m);
      mpz_init(x3); mpz_init(y3);
      mpz_init(xh_mul); mpz_init(yh_mul);
#ifdef NRESIDUE
      NResidue.init(n);
#endif
    }
  inline ~elliptic_curves() /* Destruktor */
    {
      mpz_clear(a); mpz_clear(b);
      mpz_clear(h); mpz_clear(k); mpz_clear(m);
      mpz_clear(x3); mpz_clear(y3); 
      mpz_clear(xh_mul); mpz_clear(yh_mul);
    }
 private:
  void check_curve(const mpz_t x, const mpz_t y);
  void factor_found(const mpz_t &k);
  bool add(mpz_t &xr, mpz_t &yr,
           const mpz_t x1, const mpz_t y1,
           const mpz_t x2, const mpz_t y2);
  bool mul2(mpz_t &xr, mpz_t &yr, const mpz_t x1, const mpz_t y1);
  bool mul(mpz_t &xr, mpz_t &yr,
           const mpz_t x1, const mpz_t y1,
           unsigned int K);

  void XZ_mul2(mpz_t &xr, mpz_t &zr,
               const mpz_t x1, const mpz_t z1);
  void XZ_mul2plus1(mpz_t &xr, mpz_t &zr,
                    const mpz_t Xp0, const mpz_t Zp0,
                    const mpz_t Xp1, const mpz_t Zp1,
                    const mpz_t x1, const mpz_t z1);
  void XZ_multiply(mpz_t &xr, mpz_t &zr,
                   const mpz_t x1, const mpz_t z1,
                   unsigned int K);

 public:
  void go(const int ecm_sigma, const int phase1, const int phase2);

};     


void elliptic_curves::check_curve(const mpz_t x, const mpz_t y)
{
  // prft, ob der Punkt auf der Kurve liegt, sonst: Fehlermeldung & Abbruch
  mpz_t h,x1,y1;
  mpz_init(x1); mpz_init(y1); mpz_init(h);
  mpz_powm_ui(x1,x,3,n); mpz_mul(h,a,x); mpz_mul(h,h,x); mpz_add(x1,x1,h); mpz_add(x1,x1,x);
  mpz_mod(x1,x1,n);
  mpz_powm_ui(y1,y,2,n); mpz_mul(y1,y1,b); mpz_mod(y1,y1,n);
  int fl=mpz_cmp(x1,y1);
  mpz_clear(x1); mpz_clear(y1); mpz_clear(h);
  if (fl!=0)
   {
     cerr << "Geprfter Punkt liegt nicht auf der Kurve!" << endl;
     exit(1);
   }
}

#if defined(IS_SERVER) || !defined(USE_NETWORK)

void elliptic_curves::factor_found(const mpz_t &k)
{
  cout << endl << "Faktor gefunden in elliptic curve" << endl;
  if (mpz_cmp(k,n)==0) // gefundener Faktor = n ?
   {
     cout << "Gefundene Faktorisierung ist trivial. Breche ab." << endl;
     return; // trivialer Teiler -> Verfahren abbrechen
   }
  if (mpz_probab_prime_p(k,probab_prime_checks))
    {
      cout << k << " ist Faktor." << endl;
      if (mpz_sizeinbase(k,10)<20)
       Factorization_to_file << MAL() << k << " [ecm]" << flush;
      else
       Factorization_to_file << MAL() << k << " [ecm,sigma=" << sigma << "]" << flush;
    }
  else
    {
      cout << k << " ist zusammengesetzter Faktor." << endl;
      if (mpz_sizeinbase(k,10)<20)
       Factorization_to_file << MAL() << k << " [ecm] [compound]" << flush;
      else
       Factorization_to_file << MAL() << k << " [ecm,sigma=" << sigma << "] [compound]" << flush;
    }
  mpz_div(n,n,k);
}

#else
// net-client-mode -> Faktor an den server bertragen

void elliptic_curves::factor_found(const mpz_t &k)
{
  cout << endl << "Faktor gefunden in elliptic curve" << endl;
  if (mpz_cmp(k,n)==0) // gefundener Faktor = n ?
   {
     cout << "Gefundene Faktorisierung ist trivial. Breche ab." << endl;
     return; // trivialer Teiler -> Verfahren abbrechen
   }
  if (mpz_probab_prime_p(k,probab_prime_checks))
    {
      cout << k << " ist Faktor." << endl;
    }
  else
    {
      cout << k << " ist zusammengesetzter Faktor." << endl;
    }
  mpz_div(n,n,k);

  unix_io_client communication_stream(communication_name, server_port);
  communication_stream << "Faktor(ECM)! ";
  communication_stream << sigma << " " << k << endl;

  cout << "Faktor an Server bertragen..." << endl;
}

#endif


void elliptic_curves::XZ_mul2(mpz_t &xr, mpz_t &zr,
                              const mpz_t x1, const mpz_t z1)
{
  // zu berechnen:
  // xr=(x^2-z^2)^2
  // zr=4*x*z*(x^2+A*x*z+z^2)
#if 1 // schnelle Version
  // xr=(x^2-z^2)^2=((x+z)*(x-z))^2=(x+z)^2*(x-z)^2
  // 4*x*z=(x+z)^2-(x-z)^2
  // zr/(4*x*z)=x^2+A*x*z+z^2 = (x-z)^2 + 2*x*z + a*x*z = (x-z)^2+(a+2)*x*z
  // und wenn wir b=(a+2)/4 setzen, gilt:
  // zr=4*x*z * ( (x-z)^2 + b*4*x*z )
  // ALSO WICHTIG: b=(a+2)/4 mu gegeben sein !!!
  
  mpz_add(h,x1,z1); mpz_mul(h,h,h); modulo(h,h,n); // h=(x1+z1)^2
  mpz_sub(k,x1,z1); mpz_mul(k,k,k); modulo(k,k,n); // k=(x1-z1)^2
  mpz_mul(m,h,k); modulo(xr,m,n); // xr=h*k
 
  mpz_sub(m,h,k); // m=h-k=4*x1*z1
  mpz_mul(h,b,m); modulo(h,h,n); // h=b*m
  mpz_add(k,k,h); // k=(x1-z1)^2+b*m
  mpz_mul(k,k,m); modulo(zr,k,n); // zr=4*x1*z1*k

#else// langsamere Version (getestet und funktioniert)
  // hier werden die Werte ohne Umformung ausgerechnet
  mpz_mul(h,x1,x1); modulo(h,h,n); // h=x1^2
  mpz_mul(k,z1,z1); modulo(k,k,n); // k=z1^2
  mpz_sub(m,h,k); // m=x1^2-z1^2
  mpz_add(h,h,k); // h=x1^2+z1^2
  mpz_mul(k,x1,z1); modulo(k,k,n); // k=x1*z1
  mpz_mul(m,m,m); modulo(xr,m,n); // xr=m^2
  mpz_mul(m,k,a); modulo(m,m,n); mpz_add(h,h,m); // h=k*a+x1^2+z1^2
  mpz_mul(h,h,k); mpz_mul_ui(h,h,4); modulo(zr,h,n); // zr=4*x1*z1*h
#endif
}

void elliptic_curves::XZ_mul2plus1(mpz_t &xr, mpz_t &zr,
                                   const mpz_t Xp0, const mpz_t Zp0,
                                   const mpz_t Xp1, const mpz_t Zp1,
                                   const mpz_t x1, const mpz_t z1)
{
#if 1 // schnelle Version (getestet und funktioniert)
  mpz_sub(h,Xp1,Zp1); mpz_add(m,Xp0,Zp0);
  mpz_mul(h,h,m); modulo(h,h,n); // h=(Xp1-Zp1)*(Xp0+Zp0)
  mpz_add(k,Xp1,Zp1); mpz_sub(m,Xp0,Zp0);
  mpz_mul(k,k,m); modulo(k,k,n); // k=(Xp1+Zp1)*(Xp0-Zp0)

  // im folgenden: m=(h+k)^2, h=(h-k)^2
  //               xr=z1*m,   zr=x1*h
  mpz_add(m,h,k); mpz_sub(h,h,k);
  mpz_mul(m,m,m); modulo(m,m,n); mpz_mul(m,m,z1);
  mpz_mul(h,h,h); modulo(h,h,n); mpz_mul(h,h,x1);
  modulo(xr,m,n); // xr=z1*m
  modulo(zr,h,n); // zr=x1*h

#else // langsame Version (getestet und funktioniert)
  mpz_mul(h,Xp0,Xp1); modulo(h,h,n);
  mpz_mul(k,Zp0,Zp1); modulo(k,k,n);
  mpz_sub(h,h,k); mpz_mul(h,h,h); modulo(h,h,n);
  mpz_mul(m,h,z1);

  mpz_mul(h,Zp0,Xp1); modulo(h,h,n);
  mpz_mul(k,Xp0,Zp1); modulo(k,k,n);
  mpz_sub(h,h,k); mpz_mul(h,h,h); modulo(h,h,n);
  mpz_mul(h,h,x1); modulo(zr,h,n);

  modulo(xr,m,n);
#endif
}

void elliptic_curves::XZ_multiply(mpz_t &xr, mpz_t &zr,
                                  const mpz_t x1, const mpz_t z1,
                                  unsigned int K)
{
#ifdef DEBUGMODE
  cout << "XZ_multiply " << K << " -> " << endl;
#endif
  
  bool s[100];
  short signed int i=0;

  K>>=1;
  while (K>1)
   {
     if (K&1) s[i++]=true; else { s[i++]=false; ++K; }
     K>>=1;
   }
#ifdef DEBUGMODE  
  int value=1, valuep1=2;
#endif
  mpz_t Xp0,Zp0,Xp1,Zp1;
  mpz_init_set(Xp0,x1); mpz_init_set(Zp0,z1);
  mpz_init(Xp1); mpz_init(Zp1);
  XZ_mul2(Xp1,Zp1,x1,z1);

  while (--i>=0)
   {
     if (s[i])
      {
#ifdef DEBUGMODE
        value=value+valuep1; valuep1=valuep1+valuep1;
#endif
        XZ_mul2plus1(Xp0,Zp0,Xp0,Zp0,Xp1,Zp1,x1,z1);
        XZ_mul2(Xp1,Zp1,Xp1,Zp1);
      }
     else
      {
#ifdef DEBUGMODE
        valuep1=value+valuep1; value=value+value;
#endif
        XZ_mul2plus1(Xp1,Zp1,Xp0,Zp0,Xp1,Zp1,x1,z1);
        XZ_mul2(Xp0,Zp0,Xp0,Zp0);
      }
     //cout << i << ". " << s[i] << " " << value << "," << valuep1 << endl;
   }
  
  // wenn wir davon ausgehen, da K ungerade war, gilt nun:
  // Ergebnis = value+valuep1
#ifdef DEBUGMODE
  cout << "value=" << value+valuep1 << endl;
#endif
  XZ_mul2plus1(xr,zr,Xp0,Zp0,Xp1,Zp1,x1,z1);

  mpz_clear(Xp0); mpz_clear(Zp0);
  mpz_clear(Xp1); mpz_clear(Zp1);
}

bool elliptic_curves::mul2(mpz_t &xr, mpz_t &yr, const mpz_t x1, const mpz_t y1)
{
  // !!! IMPORTANT: Don't use NResidues in this function !!!
  //cout << "";
  mpz_mul_ui(k,y1,2); mpz_mul(k,k,b); mpz_mod(k,k,n);
  if (!mpz_invert(k,k,n))
    {
      mpz_mul_ui(k,y1,2); mpz_mul(k,k,b); mpz_mod(k,k,n); mpz_gcd(k,k,n);
      factor_found(k);
      return true; // mit dieser Kurve sind wir fertig
    }
  mpz_mul(m,x1,x1); mpz_mod(m,m,n); mpz_mul_ui(m,m,3);
  mpz_mul(h,a,x1); mpz_mul_ui(h,h,2); mpz_add(m,m,h); mpz_add_ui(m,m,1);
  mpz_mul(m,m,k); mpz_mod(m,m,n);
  mpz_mul(x3,m,m); mpz_mul(x3,x3,b); mpz_sub(x3,x3,a); mpz_sub(x3,x3,x1); mpz_sub(x3,x3,x1); mpz_mod(x3,x3,n);
  mpz_sub(y3,x1,x3); mpz_mul(y3,y3,m); mpz_sub(y3,y3,y1); mpz_mod(yr,y3,n);
  mpz_set(xr,x3);
#ifdef DEBUGMODE
  cout << "mul2 " << x << "," << y << endl;
  check_curve(xr,yr);
#endif
  return false; // kein Faktor gefunden
}

bool elliptic_curves::add(mpz_t &xr, mpz_t &yr,
                          const mpz_t x1, const mpz_t y1,
                          const mpz_t x2, const mpz_t y2)
{
  // !!! IMPORTANT: Don't use NResidues in this function !!!
  if (mpz_cmp(x1,x2)) // x1!=x2
   {
     //cout << "'";
     mpz_sub(k,x2,x1); mpz_mod(k,k,n);
     if (!mpz_invert(k,k,n))
      {
        mpz_sub(k,x2,x1); mpz_mod(k,k,n); mpz_gcd(k,k,n);
        factor_found(k);
        return true; // mit dieser Kurve sind wir fertig
      }
     mpz_sub(m,y2,y1); mpz_mul(m,m,k); mpz_mod(m,m,n);
     mpz_mul(x3,m,m); mpz_mul(x3,x3,b); mpz_sub(x3,x3,a); mpz_sub(x3,x3,x1); mpz_sub(x3,x3,x2); mpz_mod(x3,x3,n);
     mpz_sub(y3,x1,x3); mpz_mul(y3,y3,m); mpz_sub(y3,y3,y1); mpz_mod(y3,y3,n);
   }
  else // x1==x2
   {
     if (mpz_cmp(y1,y2)==0) // y1==y2
       {
         //cout << ".";
         mpz_mul_ui(k,y1,2); mpz_mul(k,k,b); mpz_mod(k,k,n);
         if (!mpz_invert(k,k,n))
          {
            mpz_mul_ui(k,y1,2); mpz_mul(k,k,b); mpz_mod(k,k,n); mpz_gcd(k,k,n);
            factor_found(k);
            return true; // mit dieser Kurve sind wir fertig
          }
	 mpz_mul(m,x1,x1); mpz_mod(m,m,n); mpz_mul_ui(m,m,3);
         mpz_mul(h,a,x1); mpz_mul_ui(h,h,2); mpz_add(m,m,h);
         mpz_add_ui(m,m,1);
         mpz_mul(m,m,k); mpz_mod(m,m,n);
         mpz_mul(x3,m,m); mpz_mul(x3,x3,b); mpz_sub(x3,x3,a); mpz_sub(x3,x3,x1); mpz_sub(x3,x3,x2); mpz_mod(x3,x3,n);
         mpz_sub(y3,x1,x3); mpz_mul(y3,y3,m); mpz_sub(y3,y3,y1); mpz_mod(y3,y3,n);
       }
      else // y1!=y2
       {
         cout << "Elliptic curves: Nicht implementierter Sonderfall" << endl;
         return true;
       } 
   }

  mpz_set(xr,x3); mpz_set(yr,y3);
#ifdef DEBUGMODE
  cout << "add" << endl;
  check_curve(xr,yr);
#endif
  return false; // kein Faktor gefunden
}

bool elliptic_curves::mul(mpz_t &xr, mpz_t &yr,
                        const mpz_t x1, const mpz_t y1,
                        unsigned int K)
{
  // !!! IMPORTANT: Don't use NResidues in this function !!!

  mpz_set(xh_mul,x1); mpz_set(yh_mul,y1);

  //kleines Problem: Addition mit neutralem Element mit add nicht mglich!
  //also wird folgendermaen verfahren:
#ifdef DEBUGMODE
  if (K==0) { cerr << "Multiplikation mit Null nicht erlaubt!"; exit(1); }
#endif
  while ((K&1)==0)
   {
     K>>=1;
     if (mul2(xh_mul,yh_mul,xh_mul,yh_mul)) return true; // Faktor gefunden
   }
  // nun gilt (K&1)==1
  mpz_set(xr,xh_mul); mpz_set(yr,yh_mul);
  while (K>>=1)
   {
     if (mul2(xh_mul,yh_mul,xh_mul,yh_mul)) return true; // Faktor gefunden
     if (K&1) if (add(xr,yr,xr,yr,xh_mul,yh_mul)) return true; // Faktor gefunden
   }
#ifdef DEBUGMODE
  cout << "mul" << endl;
  check_curve(xr,yr);
#endif
  return false; // kein Faktor gefunden
}

void elliptic_curves::go(const int ecm_sigma, const int phase1, const int phase2)
{
  mpz_t x,y,z, saved_x,saved_y, xh,yh;
  mpz_init(x); mpz_init(y); mpz_init(z);
  mpz_init(saved_x); mpz_init(saved_y);
  mpz_init(xh); mpz_init(yh);
  sigma=ecm_sigma;

#if 1

  /*
     Nachfolgende Zeile aktivieren, falls das Sigma manuell
     eingegeben werden soll. (Kann eigentlich nur dann
     sinnvoll sein, wenn man zu Debuggingzwecken oder
     fr eine bekannte Zahl das Timing bestimmen will.)
  */
  //cout << "Sigma= "; cin >> sigma;

  mpz_set_ui(x,sigma); mpz_mul(x,x,x); mpz_sub_ui(x,x,5); mpz_mod(x,x,n);
  mpz_set_ui(y,sigma); mpz_mul_ui(y,y,4); mpz_mod(y,y,n);
  mpz_powm_ui(b,x,3,n); mpz_mul(b,b,y); mpz_mul_ui(b,b,4); mpz_mod(b,b,n);
  mpz_invert(a,b,n);
  mpz_sub(xh,y,x); mpz_powm_ui(xh,xh,3,n);
  mpz_mul_ui(yh,x,3); mpz_add(yh,yh,y); mpz_mul(xh,xh,yh);
  mpz_mul(a,a,xh); mpz_sub_ui(a,a,2); mpz_mod(a,a,n);

  mpz_powm_ui(x,x,3,n); mpz_powm_ui(z,y,3,n);

//#ifdef LANGSAM
  // konvertieren (nur erforderlich, falls in der Form (x:y:1)
  // gearbeitet werden soll, schadet aber auch sonst nicht viel)
  mpz_invert(z,z,n); mpz_mul(x,x,z); mpz_mod(x,x,n);
  mpz_set_ui(y,1);
  mpz_powm_ui(xh,x,3,n); mpz_powm_ui(yh,x,2,n); mpz_mod(yh,yh,a);
  mpz_add(xh,xh,yh); mpz_add(b,xh,x); mpz_mod(b,b,n);
  mpz_set_ui(z,1); // (x:y:z)=(x:1:1)
//#endif

#else
  cout << "Alternative Sigma-Methode" << endl;
  //cout << "Sigma= "; cin >> sigma;
  if (sigma==2) { cerr << "wrong sigma (must be <>2)!" << endl; return; }
  mpz_set_ui(a,sigma); mpz_set_ui(y,1);
  mpz_set_ui(x,2); 
  mpz_mul_ui(b,a,4); mpz_add_ui(b,b,10);
#endif

  cout << "Trying Elliptic Curve Method with " << endl;
  cout << "simga= " << sigma << endl;
  cout << "x=" << x << endl;
  cout << "y=" << y << endl;
  cout << "a=" << a << endl;
  cout << "b=" << b << endl;

#ifdef DEBUGMODE
  check_curve(x,y);
#endif

#ifndef LANGSAM
  // fr die XZ_mal2-Prozedur wird (in der schnellen Version)
  // b=(a+2)/4 erwartet!
  mpz_set_ui(h,4); mpz_invert(h,h,n);
  mpz_add_ui(b,a,2); mpz_mul(b,b,h); mpz_mod(b,b,n);
#endif

#if defined(NRESIDUE) && !defined(LANGSAM)
 // Die wesentlichen Parameter nach NResidue konvertieren
 NResidue.convert(a); NResidue.convert(b);
 NResidue.convert(x); NResidue.convert(y); NResidue.convert(z);
#endif

  int i,j,k,d, ausgabe_zaehler=1;
  // Phase 0: kleine Primzahlen (mit hohen Primzahlpotenzen)
  cout << "Elliptic Curve Phase 0" << endl;
#ifdef LANGSAM
  for (j=1; j<=50; j++) if (mul2(x,y,x,y)) goto fertig; // 2er-Potenzen
  for (j=1; j<=25; j++) if (mul(x,y,x,y,3)) goto fertig; // 3er-Potenzen
#else
  for (j=1; j<=50; j++) XZ_mul2(x,z,x,z); // 2er-Potenzen
  for (j=1; j<=25; j++) 
    { 
      XZ_mul2(xh,yh,x,z);
      XZ_mul2plus1(x,z,xh,yh,x,z,x,z); // 3er-Potenzen
    }
#endif
  d=4; i=1; // +4+2 Folge nutzen, um 2,3 Teilbarkeit auszuschlieen
  while (i<500)
   {
     do { i+=d; d=6-d; } while (!is_prime(i));
     j=phase2; k=1;
     while (j>=i) { k*=i; j/=i; }
#ifdef LANGSAM
     if (mul(x,y,x,y,k)) goto fertig;
     if (mul(x,y,x,y,k)) goto fertig; // ruhig zweimal...
#else
     XZ_multiply(x,z,x,z,k);
     XZ_multiply(x,z,x,z,k); // ruhig zweimal...
#endif
   }

#ifndef LANGSAM
  mpz_gcd(h,z,n); if (mpz_cmp_ui(h,1)) { factor_found(h); goto fertig; }
#endif


#ifdef DEBUGMODE
  check_curve(x,y);
#endif

  // Phase 1: Primzahlen mit (niedrigen) Potenzen
  i=i-(i%6)-1-6; d=2; // i mit Eigenschaft i=5 (mod 6) fr schnellere Primzahlermittlung
  cout << "Elliptic Curve Phase 1" << endl;
  while (i<=phase1)
    {
      do { i+=d; d=6-d; } while (i%5==0 || i%7==0 || i%11==0 || i%13==0 ||
                                 i%17==0 || i%19==0 || i%23==0 || i%29==0 ||
                  		 i%31==0 || i%37==0 || i%41==0 || i%43==0 ||
           		         i%47==0 || i%53==0 || i%59==0 || i%61==0 ||
				 powmod(2,i-1,i)!=1);
        // obige do-Schleife soll mglichst schnell viele zusammengesetzte
        // Zahlen ausschlieen, powmod(..) = Fermat-Test
      if (--ausgabe_zaehler==0)
       { // Status-Ausgabe
         ausgabe_zaehler=250;
#ifndef LANGSAM
         mpz_gcd(h,z,n); if (mpz_cmp_ui(h,1)) { factor_found(h); goto fertig; }
#endif
         cout << "Phase1: " << i << "/" << phase1 << "\r" << flush;
       }
      j=phase2; k=1;
      while (j>=i) { k*=i; j/=i; }
#ifdef LANGSAM
      if (mul(x,y,x,y,k)) goto fertig;
#else
      XZ_multiply(x,z,x,z,k);
#endif
    }
  cout << "Elliptic Curve Phase 1 bis " << i << endl;

#if 0
#ifdef LANGSAM
  cout << "Testwert (x:y:1)" << x << endl;
  { int ii; cin >> ii; }  
#else
  mpz_invert(h,z,n); mpz_mul(h,h,x); modulo(h,h,n);
  cout << "Testwert (x::z)" << h << endl;
  { int ii; cin >> ii; }  
#endif
#endif

#ifndef LANGSAM
  // noch einmal auf Teiler testen...
  mpz_gcd(h,z,n); if (mpz_cmp_ui(h,1)) { factor_found(h); goto fertig; }
  mpz_gcd(h,x,n); if (mpz_cmp_ui(h,1)) { factor_found(h); goto fertig; }

#if defined(NRESIDUE) && !defined(LANGSAM)
 // Die wesentlichen Parameter von NResidue in die Normalform konvertieren
 NResidue.convert_back(a); NResidue.convert_back(b);
 NResidue.convert_back(x); NResidue.convert_back(y); NResidue.convert_back(z);
#endif

  // und nun Rckkonvertierung in die (x:y:1)-Form
  mpz_set_ui(y,1);
  mpz_invert(h,z,n); mpz_mul(x,x,h); mpz_mod(x,x,n); // x=x/z

  mpz_powm_ui(xh,x,3,n); mpz_powm_ui(yh,x,2,n); mpz_mul(yh,yh,a);
  mpz_add(xh,xh,yh); mpz_add(b,xh,x); mpz_mod(b,b,n);
  mpz_set_ui(z,1); // (x:y:z)=(x:1:1)

  check_curve(x,y);
#endif


#ifndef ECM_FFT_CONTINUATION

/* IMPORTANT NOTE:
   In phase 2 never use NResidues for calculating points on the curve.
   But it should be worthwhile - if NRESIDUE is defined - to work with
   "bfact" (and a transformed copy of "x" named "saved_x") as NResidues
   to speed up the modular multiplications with "sammel".
*/

  {
   // Phase 2: isolierte (Prim-)zahl: Kurve mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Elliptic Curve Phase 2 (improved)" << endl;
   mpz_set(saved_x,x); mpz_set(saved_y,y); // Basis fr Phase2

   // precomputing most common used differences:
   // pre_mul_size ist die Intervallgre, in der auch gesiebt wird
   int pre_mul_size = 2*3*5; // 2*3 ist mindestens erforderlich!
   {
     // Je mehr verschiedene, kleine Primfaktoren enthalten sind, desto
     // weniger teilerfremde Zahlen gibt es im Intervall -> desto geringer
     // sind also die Speicherkosten.
     // Ideale Intervallgre wre etwa sqrt(phase2), wenn man die Kosten
     // fr die Ermittlung der Primzahlen, usw. vernachlssigt.
     int p=5;
     do
      { 
        do p+=2; while (!is_prime(p));
        pre_mul_size*=p;
      } while (pre_mul_size<=phase2/pre_mul_size);

     if (pre_mul_size>3*sqrt(phase2)) // Schwellwert-Heuristik
      {
        pre_mul_size/=p; // grter Primfaktor war zu gro -> abdividieren
        // stattdessen noch bis zur optimalen Gre mit Zusatzfaktor auffllen:
        pre_mul_size*=static_cast<int>(ceil(sqrt(phase2)/pre_mul_size));
      }
     cout << "pre_mul_size=" << pre_mul_size << endl;
   }

   mpz_t bfact[pre_mul_size]; // hier wird der zu einer Differenz gehrige
                              // x-Kurvenwert abgelegt

   /* Im folgenden wird die Tabelle bfact aufgebaut, die von dem erreichten
      Startpunkt P=(saved_x:saved_y) die Punkte k*P fr 0<=k<pre_mul_size
      enthlt. Genauer: bfact[k]=X-Wert von k*P (den Y-Wert bentigen wir
      fr sptere Zugriffe nicht mehr).
      Die Tabelle wir spter bentigt, um in pre_mul_size-Schritten auf der
      elliptischen Kurve voranspringen zu knnen. Einen Punkt m*P knnen
      wir dann erreichen, in dem wir m1=floor(m/pre_mul_size), m2=m%premulsize
      setzen und m*P=(m1*premulsize)*P+m2*P berechnen.
      Fr einzelne Wertberechnungen bringt uns diese Form nichts, aber fr
      viele sukzessive Werte reicht es dann, den Punkt Q=premulsize*P
      zu berechnen, womit m*P=m1*Q+m2*P gilt.
      Wir haben nun erreicht, da
        1. m2*P =:R[m2] fr 0<=m2<pre_mul_size ber Tabellenzugriffe geholt
           werden kann. -> "kostenlos"
        2. m1*premulsize*P=m1*Q =:S ist fr Werte innerhalb eines Intervalls
           der Gre premulsize ebenfalls konstant, solange m1 sich nicht
           ndert!
       -> Die Kosten reduzieren sich innerhalb eines Intervalls auf
          die Kosten der Addition von S+R[m2].
     
      Da wir in der improved-standard-continuation fr k lediglich alle
      aufeinanderfolgenden Primzahlen (die grer sind als der letzte Wert
      aus Phase 1) durchlaufen wollen, knnen wir fr premulsize das
      Produkt kleiner Primzahlen whlen und brauchen fr m2 nur die zu
      premulsize teilerfremden Zahlen zu nehmen.
      Um diese Punkte R[m2] an diesen Punkten zu berechnen, empfiehlt es sich
      fr groes premulsize, die Punkte hnlich der standart-continuation
      zu ermitteln; d.h. in Differenzen voranzuspringen und die Punkte
      ber Additionen der (vorberechneten) "Differenzpunkte" zu bestimmen.

      Der vorangegangene Kommentar beschreibt zwar nicht exakt unsere
      tatschliche Vorgehensweise, aber er umschreibt die dahinterliegende
      "Idee", die umgesetzt wird...
   */
      
   const int table_dxy_size = 23;
   mpz_t table_dxy[table_dxy_size][2];

   if (mul2(x,y,saved_x,saved_y)) goto fertig; // Primzahldifferenzen sind gerade (fr p>3)
   mpz_set(xh,x); mpz_set(yh,y); // (x,y) nach (xh,yh) sichern
 
   mpz_init_set(table_dxy[2][0],x);
   mpz_init_set(table_dxy[2][1],y);
   for (int j=4; j<table_dxy_size; j+=2) 
    {
      if (add(x,y,x,y,xh,yh)) goto fertig;
      mpz_init_set(table_dxy[j][0],x);
      mpz_init_set(table_dxy[j][1],y);
    }

   mpz_set(x,xh); mpz_set(y,yh); // (x,y) zurckholen
   for (int j=pre_mul_size-1, jlast=pre_mul_size-1; j>=0; j-=2)
    {
      if (gcd(j,pre_mul_size)==1) // j teilerfremd?
       {
         /* 
            aktuellen Punkt (premulsize-1-j)*(xh,yh) berechnen und
            die X-Ordinate in bfact[j] ablegen
         */
         int dj=jlast-j;
         if (dj>=table_dxy_size)
          { 
            cerr << "increase table_dxy_size to " << dj+1 << "!" << endl;
            exit(1);
          }
         if (dj>0) // Sonderfall: fr dj=0 nicht addieren
          if (add(x,y,x,y,table_dxy[dj][0],table_dxy[dj][1])) goto fertig;
         jlast=j;
         mpz_init_set(bfact[j],x);
#ifdef NRESIDUE
         NResidue.convert(bfact[j]);
#endif
       }
    }

   for (int j=2; j<table_dxy_size; j+=2) // Tabelle der Differenzpunkte lschen
    {
      mpz_clear(table_dxy[j][0]);
      mpz_clear(table_dxy[j][1]);
    }


   int next_i;
   mpz_t sammel;
   mpz_init(sammel); mpz_set_ui(sammel,1);

   // vielleicht etwas umstndlich, aber dafr verstndlich:
   if (mul(xh,yh,saved_x,saved_y,pre_mul_size)) goto fertig; // Intervall-Summand
   // (xh,yh)=premulsize*(saved_x,saved_y) ist nun der Intervallsummand

   i=((i/pre_mul_size)*pre_mul_size)-1; d=2; // i mit Eigenschaft i=5 (mod 6) fr schnellere Primzahlermittlung
   if (mul(x,y,saved_x,saved_y,i)) goto fertig; // Startwert

   next_i = i+1+pre_mul_size; // Intervallgrenze setzen

   bool sieve[pre_mul_size];
   for (int i1=pre_mul_size-1; i1>=0; i1-=2) sieve[i1]=false;

   for (int p=5, dp=2; p<pre_mul_size && p*p<next_i; p+=dp, dp=6-dp )
    for (int i1=p-((next_i-pre_mul_size)%p); i1<pre_mul_size; i1+=p) sieve[i1]=true;

   while (i<=phase2) 
    { 
      int runden=25000; // alle "runden" Runden mit gcd auf Faktor testen
      while (runden>0)
	{
          // nchste Primzahl ermitteln...
          do { i+=d; d=6-d; } while (i<next_i && sieve[i%pre_mul_size]);

         /* falls pre_mul_size gro genug gewhlt wurde, wre statt
            "while"-Schleife auch eine "if"-Bedingung ausreichend... */
         while (i>next_i) // neues Intervall berechnen?
          { 
            next_i += pre_mul_size;
            if (add(x,y,x,y,xh,yh)) goto fertig_phase2;
#ifdef NRESIDUE
            mpz_set(saved_x,x);
            NResidue.convert(saved_x);
#endif
            for (int i1=pre_mul_size-1; i1>=0; i1-=2) sieve[i1]=false;
            for (int p=5, dp=2; p<pre_mul_size && p*p<next_i; p+=dp, dp=6-dp)
             for (int i1=p-((next_i-pre_mul_size)%p); i1<pre_mul_size; i1+=p) sieve[i1]=true;
          }

          // nun in "sammel" den mglicherweise gefundenen Faktor sammeln... 
#ifdef NRESIDUE
          mpz_sub(h,saved_x,bfact[next_i-i]);
#else
          mpz_sub(h,x,bfact[next_i-i]);
#endif        
          mpz_mul(sammel,sammel,h); modulo(sammel,sammel,n);
	  --runden;
	}

      /* Bemerkung:
         Wenn wir das in "sammel" gebildete Produkt als Auswertung
         eines Polynoms (x-c1)*(x-c2)*...*(x-ck) auffassen, dann
         bietet es sich an, eine Polynomauswertung zu implementieren, die
         schneller arbeitet als die k-fache Multiplikation.

         Folgende Vorgehensweise wre dann zu implementieren:
          1. Nullstellenform des Polynoms ermitteln
             -> P(x)=(x-c1)*(x-c2)*...*(x-ck)
          2. P(x) ausmultiplizieren
             -> P(x)=a0 + a1*x + a2*x^2 + ... + a[k-1]*x^(k-1) + x^k
          3. k Werte (x1 ... x[k]) aufsammeln, fr die P(x) ausgewertet
             werden soll (-> k Intervallwerte von x bestimmen) 
          4. r[i]=P(x[i]) fr i=1..k bestimmen,
             z.B. mit schneller Auswertung ber
             r[i]=P(x) modulo (x-x[i]) fr i=1...k simultan auswerten.
             (-> divide and conquer)
          5. Produkt R=r1*r2*...r[k] modulo N berechnen
             und gcd(R,N) bestimmen. 
          6. Falls Schritt 5 keinen Teiler geliefert hat und die
             Obergrenze nicht erreicht ist, ab Schritt 3 fortfahren.

         Bemerkung: Um gengend viele gleiche Polynome auszuwerten zu knnen,
           sollte man hierbei auf Primzahltests verzichten und immer
           alle Nullstellen auswerten. (-> Nur ein Polynom bestimmen.)
           Mit asymptotisch schnellen Algorithmen sollte dies dann in etwa
           der FFT-continuation entsprechen.
           In der konkreten Umsetzung mte es allerdings ausreichen, den
           Punkt 4 einigermaen effizient zu implementieren.
      */

      // und erst nach "runden"-vielen Durchlufen die aufgesammelten Zahlen
      // auf grten gemeinsamen Teiler testen...
      mpz_gcd(sammel,sammel,n);
      if (mpz_cmp_ui(sammel,1)) { factor_found(sammel); goto fertig_phase2; }
      mpz_set_ui(sammel,1); // sptere Multiplikation erleichtern...
      cout << "Phase2: " << i << "/" << phase2 << "\r" << flush;
    }

 fertig_phase2:
  mpz_clear(sammel);
  cout << "Elliptic Curve Phase 2 bis " << i << endl;
   for (int j=pre_mul_size-1; j>=0; j-=2)
    {
      // freeing memory for precomputed differences
      // !! siehe auch die Allokierung!
      if (gcd(j,pre_mul_size)==1) mpz_clear(bfact[j]);
    }  
  }

// end of ecm improved standard continuation

#else
 //---------------------------------------------------------------
 // ecm fft-continuation
 //---------------------------------------------------------------

/* WARNUNG: diese Implementierung ist noch nicht sonderlich ausgereift...
   Es ist nur eine fr fft-continuation umgemodelte Version der
   vorangegangenen improved standard continuation.
   fft bedeutet in diesem Zusammhang, da multipoint polynomial evaluation
   verwendet wird. Diese baut auf dem Prinzip der Fast-Fourier-Transformation auf.
   Fr die konkrete Implementierung siehe das Modul "polynomial.cc".
   Die eigentliche Optimierungsarbeit mu in dem oben genannten Modul vorgenommen werden.
   In spteren Versionen wird sicherlich auch noch eine auf der FFT basierende
   Polynommultiplikation implementiert werden.
*/


/* IMPORTANT NOTE:
   In phase 2 (fft) never use NResidues for calculating points on the curve.
*/

  {
   // Phase 2: isolierte (Prim-)zahl: Kurve mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Elliptic Curve Phase 2 (fft)" << endl;
   mpz_set(saved_x,x); mpz_set(saved_y,y); // Basis fr Phase2

   // precomputing most common used differences:
   // pre_mul_size ist die Intervallgre mit der pro Polynom gesucht wird

     // Je mehr verschiedene, kleine Primfaktoren enthalten sind, desto
     // weniger teilerfremde Zahlen gibt es im Intervall -> desto geringer
     // sind also die Speicherkosten.
     // bei unserer fft-Version gilt in etwa, da phi*pre_mul_size ungefhr phase2
     // entsprechen sollte; d.h. das Polynom hat die Grenordnung phi und
     // greift ein Intervall der Gre pre_mul_size ab. Und #phi Funktionswerte lassen
     // sich dann schnell ber multipoint_eval ermitteln.
     // Nun ist die Polynomauswertung allerdings besonders effizient, wenn die
     // Polynome und die Anzahl der Punkte Zweierpotenzen sind.
     // Meiner Ansicht nach drfte es daher vorteilhaft sein, pre_mul_size so zu
     // whlen, da es maximal wird fr ein phi, welches nahe unterhalb der
     // gewnschten Zweierpotenz liegt.
     // Da es nun nicht allzu viele praktikable Zweierpotenzen gibt, mit denen
     // wir arbeiten knnen (bei kleinen Zweipotenzen ist die improved standard
     // continuation berlegen, bei zu groen Zweierpotenzen ist der Speicherbedarf
     // zu hoch), knnen wir geeignete Werte ber eine Tabelle vorgeben.
     /* Vorschlge wren 
        Polynomgrad  phi(pre_mul_size) 	pre_mul_size                	abgedeckter Suchraum
	2^8 =    256,  	   240,		 1050=(2)*(3)*(5)^2*(7),		     134.400
	2^9 =    512,  	   480,		 2310=(2)*(3)*(5)*(7)*(11),		     591.360
	2^10=   1024,  	   960,	 	 4620=(2)^2*(3)*(5)*(7)*(11),		   2.365.440
	2^11=   2048,	  1920,	 	 9240=(2)^3*(3)*(5)*(7)*(11),		   9.461.760
	2^12=   4096,	  4032,	 	19110=(2)*(3)*(5)*(7)^2*(13),		  39.137.280
	2^13=   8192,	  7680,	 	39270=(2)*(3)*(5)*(7)*(11)*(17),         160.849.920
	2^14=  16384,    16128,	        79170=(2)*(3)*(5)*(7)*(13)*(29),         648.560.640
	2^15=  32768,    31680,	       159390=(2)*(3)^2*(5)*(7)*(11)*(23),     2.611.445.760
	2^16=  65536,    63360,	       330330=(2)*(3)*(5)*(7)*(11)^2*(13),    10.824.253.440
	2^17= 131072,	126720,	       690690=(2)*(3)*(5)*(7)*(11)*(13)*(23), 45.265.059.840
     */

     int pg=256, pms=1050;
     if (phase2 >    591360/4*3) { pg=  512; pms=  2310; }
     if (phase2 >   2365440/4*3) { pg= 1024; pms=  4620; }
     if (phase2 >   9461760/4*3) { pg= 2048; pms=  9240; }
     if (phase2 >  39137280/4*3) { pg= 4096; pms= 19110; }
     if (phase2 > 160849920/4*3) { pg= 8192; pms= 39270; }
     if (phase2 > 648560640/4*3) { pg=16384; pms= 79170; }
     //if (phase2 > 2611445760/4*3) { pg=32768; pms=330330; }

   /* Am Rande:
      Im Augenblick scheint es sich noch nicht zu lohnen,
      die Schleifen von "int" auf "unsigned int" oder "double" umzustellen.
      Bei 64-Bit-Rechnern drfte "int" dann wieder gengend gro sein.
      Kleinere Rechner drften eh Speicherprobleme bei derart groen
      Polynomen bekommen.
   */

   const int pre_mul_size = pms;
   const int Polynom_Index = pg;

   cout << "Polynomgre=" << Polynom_Index << endl;
   cout << "Einzel-Intervallsize=" << pre_mul_size << endl; 

   mpz_t sammel[Polynom_Index];
   for (int j=0; j<Polynom_Index; ++j) mpz_init(sammel[j]);

   /*
      Fr den Aufbau des Polynoms:
      Nullstellen des Polynoms entsprechen der bfact-Tabelle der
      improved standard continuation, siehe dortige Kommentare
   */
   const int table_dxy_size = 23;
   mpz_t table_dxy[table_dxy_size][2];

   if (mul2(x,y,saved_x,saved_y)) goto fertig; // Primzahldifferenzen sind gerade (fr p>3)
   mpz_set(xh,x); mpz_set(yh,y); // (x,y) nach (xh,yh) sichern
 
   mpz_init_set(table_dxy[2][0],x);
   mpz_init_set(table_dxy[2][1],y);
   for (int j=4; j<table_dxy_size; j+=2) 
    {
      if (add(x,y,x,y,xh,yh)) goto fertig;
      mpz_init_set(table_dxy[j][0],x);
      mpz_init_set(table_dxy[j][1],y);
    }

   int Polynom_Index_i = 0;
   mpz_set(x,xh); mpz_set(y,yh); // (x,y) zurckholen
   for (int j=pre_mul_size-1, jlast=pre_mul_size-1; j>=0; j-=2)
    {
      if (gcd(j,pre_mul_size)==1) // j teilerfremd?
       {
         /* 
            aktuellen Punkt (premulsize-1-j)*(xh,yh) berechnen und
            die X-Ordinate in bfact[j] ablegen
         */
         int dj=jlast-j;
         if (dj>=table_dxy_size)
          { 
            cerr << "increase table_dxy_size to " << dj+1 << "!" << endl;
            exit(1);
          }
         if (dj>0) // Sonderfall: fr dj=0 nicht addieren
          if (add(x,y,x,y,table_dxy[dj][0],table_dxy[dj][1])) goto fertig;
         jlast=j;
         //mpz_init_set(bfact[j],x); Polynom_Index_i++;
         mpz_set(sammel[Polynom_Index_i++],x);
       }
    }

   for (int j=2; j<table_dxy_size; j+=2) // Tabelle der Differenzpunkte lschen
    {
      mpz_clear(table_dxy[j][0]);
      mpz_clear(table_dxy[j][1]);
    }

   cout << "unvergrerte Polynomgre: " << Polynom_Index_i << endl;

   polynomial::TPolynom Polynom; // hier steht das anschlieend konstruierte Polynom drin
 
   // Rest bis zur nchsten Zweierpotenz mit Nullen auffllen
   /* Bemerkung: Besser wre es, Punkte auerhalb des Suchbereichs
                 zu whlen (Motto: doppeltes, aber auerhalb des Suchbereichs
                 lchriges Netz)
    */
   { 
     int anz_Punkte=Polynom_Index_i; 
     while (anz_Punkte<Polynom_Index) mpz_set_ui(sammel[anz_Punkte++],0);
     Polynom_Index_i=polynomial::Polynom_aus_Nullstellen_konstruieren(Polynom,sammel,anz_Punkte,n);
     cout << "Polynom erzeugt. Polynomindex = " << Polynom_Index_i << endl;
   }
   

   // vielleicht etwas umstndlich, aber dafr verstndlich:
   if (mul(xh,yh,saved_x,saved_y,pre_mul_size)) goto fertig; // Intervall-Summand
   // (xh,yh)=premulsize*(saved_x,saved_y) ist nun der Intervallsummand

   i=((i/pre_mul_size)*pre_mul_size)-1; d=2; // i mit Eigenschaft i=5 (mod 6) fr schnellere Primzahlermittlung
   if (mul(x,y,saved_x,saved_y,i)) goto fertig; // Startwert

   int next_i = i+1+pre_mul_size; // Intervallgrenze setzen

   while (i<=phase2)
    { 

      cout << "Phase2: " << i << "/" << phase2 << endl << flush;
      cout << "Sammle Werte..." << endl;

      //Werte sammeln
      int sammel_index=0;
      while (sammel_index<(Polynom_Index_i-1)/2)
       {
         mpz_set(sammel[sammel_index++],x);
         // und neues Intervall berechnen
         i += pre_mul_size;
         next_i += pre_mul_size; 
         if (add(x,y,x,y,xh,yh)) goto fertig_phase2;
       }

      // Polynomauswertung starten
      cout << "Phase2: " << i << "/" << phase2 << endl << flush;
      cout << "starte Polynomauswertung" << endl;
      cout << "Parameter: Polynomindex: " << Polynom_Index_i
           << ", Anzahl der Punkte: " << sammel_index << endl;
      polynomial::multipoint_eval(sammel,Polynom,Polynom_Index_i,sammel,sammel_index,n);
      cout << "Polynomauswertung beendet. Teste auf ggT." << endl;
      mpz_set(h,sammel[0]);
      for (int j=1; j<sammel_index; ++j) { mpz_mul(h,h,sammel[j]); mpz_mod(h,h,n); }
      mpz_gcd(saved_x,h,n);
      /* Hinweis: Backtracking wre mglich, falls
                  Mehrfachfaktor entdeckt werden sollte:
                  a) gcd(sammel[j],n) getrennt berechnen
                  b) Polynomnullstellen einzeln untersuchen, falls auch dies scheitert.
                  Beides geschieht hier nicht...
      */   
      if (mpz_cmp_ui(saved_x,1)) { factor_found(saved_x); goto fertig_phase2; }

      /* Bemerkung:
         Wenn wir das in "sammel" gebildete Produkt als Auswertung
         eines Polynoms (x-c1)*(x-c2)*...*(x-ck) auffassen, dann
         bietet es sich an, eine Polynomauswertung zu implementieren, die
         schneller arbeitet als die k-fache Multiplikation.

         Folgende Vorgehensweise ist hier (in etwa) implementiert
          1. Nullstellenform des Polynoms ermitteln
             -> P(x)=(x-c1)*(x-c2)*...*(x-ck)
          2. P(x) ausmultiplizieren
             -> P(x)=a0 + a1*x + a2*x^2 + ... + a[k-1]*x^(k-1) + x^k
          3. k Werte (x1 ... x[k]) aufsammeln, fr die P(x) ausgewertet
             werden soll (-> k Intervallwerte von x bestimmen) 
          4. r[i]=P(x[i]) fr i=1..k bestimmen,
             z.B. mit schneller Auswertung ber
             r[i]=P(x) modulo (x-x[i]) fr i=1...k simultan auswerten.
             (-> divide and conquer)
          5. Produkt R=r1*r2*...r[k] modulo N berechnen
             und gcd(R,N) bestimmen. 
          6. Falls Schritt 5 keinen Teiler geliefert hat und die
             Obergrenze nicht erreicht ist, ab Schritt 3 fortfahren.

         Bemerkung: Um gengend viele gleiche Polynome auswerten zu knnen,
           sollte man hierbei auf Primzahltests verzichten und immer
           alle Nullstellen auswerten. (-> Nur ein Polynom bestimmen.)
           Mit asymptotisch schnellen Algorithmen sollte dies dann in etwa
           der FFT-continuation entsprechen.
           In der konkreten Umsetzung mte es allerdings ausreichen, den
           Punkt 4 einigermaen effizient zu implementieren.
      */

    }

 fertig_phase2:
  for (int j=0; j<Polynom_Index; ++j) mpz_clear(sammel[j]);
  for (int j=0; j<Polynom_Index_i; ++j) mpz_clear(Polynom[j]);
  cout << "Elliptic Curve Phase 2 bis " << i << endl;
  }

// end of ecm fft continuation



#endif


fertig:
  mpz_clear(x); mpz_clear(y); mpz_clear(z);
  mpz_clear(saved_x); mpz_clear(saved_y);
  mpz_clear(xh); mpz_clear(yh);
}
